/* 
   masq.c - masquerading various login versions

   Copyright (C) Szymon Acedanski 2003
*/

/*
    ggrelay - Gadu-Gadu instant messager relaying agent
    Copyright (C) 2003 Szymon Acedanski

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <pthread.h>

#include "ggrelay.h"

void gg_masq_hdr(struct gg_fd* f);
void gg_masq_dcc(struct gg_fd* f);

struct gg_hashtbl* gg_masq_uinh;
int gg_uinh_count = 0;
LIST_HEAD(gg_masq_cexps);

/* Helpers */

static int gg_uinh_fn(struct gg_uinh_entry* e, int size)
{
	return e->uin % size;
}

static int gg_uinh_cmp(struct gg_uinh_entry* a, struct gg_uinh_entry* b)
{
	return a->uin - b->uin;
}

static void gg_uinh_clear(void)
{
	int i;
	for (i=0; i<gg_masq_uinh->size; i++)
	{
		struct list_head* h = &gg_masq_uinh->heads[i];
		while (!list_empty(h))
		{
			struct list_head* item = h->next;
			list_del(item);
			free(item);
		}
	}
	gg_uinh_count = 0;
}

static void gg_uinh_update(uint32_t uin, uint32_t ip, uint16_t port,
		struct gg_fd* fd)
{
	struct gg_uinh_entry e;
	struct gg_uinh_entry* f;
	if (!ip) return;
	uin &= 0x00ffffff;
	e.uin = uin;
	f = gg_find_hitem(gg_masq_uinh, &e);
	DEBUGP("uin=%d ip=%u.%u.%u.%u port=%hd", uin, NIPQUAD(ip), port);
	if (f)
	{
		if (f->fd && !fd) return;
		f->ip = ip;
		f->port = port;
		f->fd = fd;
	}
	else
	{
		if (gg_uinh_count >= gg_uinh_limit)
		{
			gg_warn("warning: uins hashtable limit exceeded, clearing");
			gg_uinh_clear();
		}
		if (!(f = malloc(sizeof(e))))
		{
			gg_error("warning: cannot allocate hashtable entry");
			return;
		}
		INIT_LIST_HEAD(&f->list);
		f->uin = uin;
		f->ip = ip;
		f->port = port;
		f->fd = fd;
		gg_add_hitem(gg_masq_uinh, f);
		gg_uinh_count++;
	}
}

static inline void gg_uinh_del(uint32_t uin)
{
	struct gg_uinh_entry e;
	struct gg_uinh_entry* f;
	uin &= 0x00ffffff;
	e.uin = uin;
	f = gg_find_hitem(gg_masq_uinh, &e);
	if (f)
	{
		gg_del_hitem(gg_masq_uinh, f);
		free(f);
		gg_uinh_count--;
	}
}

static inline struct gg_uinh_entry* gg_uinh_find(uint32_t uin)
{
	struct gg_uinh_entry e;
	uin &= 0x00ffffff;
	e.uin = uin;
	return gg_find_hitem(gg_masq_uinh, &e);
}

int gg_masq_init(void)
{
	if (!(gg_masq_uinh = gg_alloc_hashtbl(gg_uinh_size,
					(int (*)(void*, int))gg_uinh_fn,
					(int (*)(void*, void*))gg_uinh_cmp)))
	{
		gg_error("fatal: error allocating uin hashtable");
		return 1;
	}
	return 0;
}

static void gg_masq_cont(struct gg_fd* f, int nextpartlen,
		void (*fn)(struct gg_fd*))
{
	gg_assert(f->writeoffset == 0);
	f->expectlen += nextpartlen;
	f->expectfn = fn;
}

static void gg_masq_nextpkt(struct gg_fd* f)
{
	gg_assert(f->writeoffset == 0);
	if (f->injectlen && f->expectlen == f->readoffset &&
			f->bufsize - f->readoffset >= f->injectlen)
	{
		/* TODO: check this code once again */
		memcpy(f->buffer + f->readoffset, f->injectbuf, f->injectlen);
		f->readoffset += f->injectlen;
	}
	DEBUGP(":)");
	f->writeoffset = f->expectlen;
	f->expectlen = sizeof(struct gg_header);
	f->expectfn = gg_masq_hdr;
}

void gg_masq_init_chat_req(struct gg_fd* f)
{
	f->expectlen = sizeof(struct gg_header);
	f->expectfn = gg_masq_hdr;
	f->state = GG_S_EXPECTING;
}

void gg_masq_init_chat_resp(struct gg_fd* f)
{
	f->expectlen = sizeof(struct gg_header);
	f->expectfn = gg_masq_hdr;
	f->state = GG_S_EXPECTING;
}

void gg_masq_init_dcc_req(struct gg_fd* f)
{
	f->expectlen = sizeof(struct gg_dcc_handshake);
	f->expectfn = gg_masq_dcc;
	f->state = GG_S_EXPECTING;
}

static void gg_proto_violated(struct gg_fd* f)
{
	gg_condlog("protocol violation on fd %d, entering repeatmode",
			f->fd);
	f->writeoffset = f->readoffset;
	f->state = GG_S_REPEATMODE;
	if (f->twin && f->twin->state != GG_S_CLOSING &&
			f->twin->state !=GG_S_DESTROY)
	{
		f->twin->writeoffset = f->twin->readoffset;
		f->twin->state = GG_S_REPEATMODE;
	}
}

static int gg_masq_realloc_buffer(struct gg_fd* f, int newsize)
{
	char* newbuf = (char*)realloc(f->buffer, newsize);
	if (!newbuf)
	{
		gg_error("warning: cannot grow buffer (fd=%d newsize=%d)",
				f->fd, newsize);
		return 1;
	}
	f->buffer = newbuf;
	f->bufsize = newsize;
	return 0;
}

static inline int gg_max(int a, int b)
{
	return a > b ? a : b;
}

/* Handlers */

void gg_masq_close_handler(struct gg_fd* f)
{
	if (f->uin)
		gg_uinh_del(f->uin);
}

/* Packet handlers */

static void gg_masq_login(struct gg_fd* f)
{
	struct gg_login* login = (struct gg_login*)(f->buffer +
			sizeof(struct gg_header));

	if (!login->local_ip)
		goto out;

	f->nat_ip = login->local_ip;
	f->nat_port = login->local_port;
	f->uin = login->uin;
	gg_uinh_update(f->uin, f->nat_ip, f->nat_port, f);

	/* packet mangling here :) */

	if (!gg_need_dcc_server())
	{
		login->local_ip = gg_outer_ip;
		login->local_port = gg_dcc_port;
	}
	
out:
	gg_condlog("caught login uin=%lu (fds %d:%d) [%s]", (unsigned long)login->uin,
			f->fd, f->twin->fd, f->uin ? "HASDCC" : "NODCC");

	gg_masq_nextpkt(f);
}

static void gg_masq_login_ext(struct gg_fd* f)
{
	struct gg_login_ext* login = (struct gg_login_ext*)(f->buffer +
			sizeof(struct gg_header));

	if (login->local_ip)
	{
		f->nat_ip = login->local_ip;
		f->nat_port = login->local_port;
	}
	else
		goto out;

	f->uin = login->uin;
	gg_uinh_update(f->uin, f->nat_ip, f->nat_port, f);

	/* packet mangling here :) */

	if (!gg_need_dcc_server())
	{
		login->external_ip = gg_outer_ip;
		login->external_port = gg_dcc_port;
	}
	
out:
	gg_condlog("caught login_ext uin=%lu (fds %d:%d) [%s]", (unsigned long)login->uin,
			f->fd, f->twin->fd, f->uin ? "HASDCC" : "NODCC");

	gg_masq_nextpkt(f);
}

static void gg_masq_login60(struct gg_fd* f)
{
	struct gg_login60* login = (struct gg_login60*)(f->buffer +
			sizeof(struct gg_header));

	if (login->d1 != 0x00)
		goto failure;
	if (login->d2 != 0xbe)
		goto failure;

	if (login->local_ip)
	{
		f->nat_ip = login->local_ip;
		f->nat_port = login->local_port;
	}
	else
		goto logfailure;

	f->uin = login->uin;
	gg_uinh_update(f->uin, f->nat_ip, f->nat_port, f);

	/* packet mangling here :) */

	if (!gg_need_dcc_server())
	{
		login->external_ip = gg_outer_ip;
		login->external_port = gg_dcc_port;
	}
	
logfailure:
	gg_condlog("caught login60 uin=%lu (fds %d:%d) [%s]", (unsigned long)login->uin,
			f->fd, f->twin->fd, f->uin ? "HASDCC" : "NODCC");

failure:
	gg_masq_nextpkt(f);
}

static void gg_masq_notify_reply(struct gg_fd* f)
{
	struct gg_notify_reply* nr = (struct gg_notify_reply*)
		(f->buffer+sizeof(struct gg_header));

	if (gg_leave_local)
	{
		struct gg_uinh_entry* h = gg_uinh_find(nr->uin);
		if (h && h->fd)
		{
			DEBUGP("friend %d is local (%u.%u.%u.%u:%d)",
					nr->uin & 0x00ffffff,
					NIPQUAD(h->ip), h->port);
			nr->remote_ip = h->ip;
			nr->remote_port = h->port;
			goto cont;
		}
	}
	gg_uinh_update(nr->uin, nr->remote_ip, nr->remote_port, NULL);
	nr->remote_ip = gg_outer_ip;
	nr->remote_port = gg_dcc_port;
		
cont:
	gg_masq_nextpkt(f);
}

static void gg_masq_notify_reply60(struct gg_fd* f)
{
	struct gg_header* hdr = (struct gg_header*)f->buffer;
	unsigned char* iter = f->buffer + sizeof(struct gg_header);
	unsigned char* limit = iter + hdr->length -
		sizeof(struct gg_notify_reply60);

	while (iter <= limit)
	{
		struct gg_notify_reply60* nr = (struct gg_notify_reply60*)iter;
		if (gg_leave_local)
		{
			struct gg_uinh_entry* h = gg_uinh_find(nr->uin);
			if (h && h->fd)
			{
				DEBUGP("friend %d is local (%u.%u.%u.%u:%d)",
						nr->uin & 0x00ffffff,
						NIPQUAD(h->ip), h->port);
				nr->remote_ip = h->ip;
				nr->remote_port = h->port;
				goto cont;
			}
		}
		gg_uinh_update(nr->uin, nr->remote_ip, nr->remote_port, NULL);
		nr->remote_ip = gg_outer_ip;
		nr->remote_port = gg_dcc_port;
cont:
		iter += sizeof(struct gg_notify_reply60);
		if (GG_S_D(nr->status))
		{
			if (iter >= limit+sizeof(struct gg_notify_reply60))
			{
				gg_proto_violated(f);
				return;
			}
			iter += *iter + 1;
		}
	}

	gg_masq_nextpkt(f);
}

static void gg_masq_status60(struct gg_fd* f)
{
	struct gg_status60* st = (struct gg_status60*)(f->buffer +
			sizeof(struct gg_header));

	if (gg_leave_local)
	{
		struct gg_uinh_entry* h = gg_uinh_find(st->uin);
		if (h && h->fd)
		{
			DEBUGP("friend %d is local (%u.%u.%u.%u:%d)",
					st->uin & 0x00ffffff,
					NIPQUAD(h->ip), h->port);
			st->remote_ip = h->ip;
			st->remote_port = h->port;
			goto cont;
		}
	}

	gg_uinh_update(st->uin, st->remote_ip, st->remote_port, NULL);
	st->remote_ip = gg_outer_ip;
	st->remote_port = gg_dcc_port;

cont:
	gg_masq_nextpkt(f);
}

static void gg_masq_default(struct gg_fd* f)
{
	gg_masq_nextpkt(f);
}

#define GG_MASQ_PKT(num, type, fn) \
	case num: \
		if (hdr->length < sizeof(type)) \
			goto failure; \
		gg_masq_cont(f, hdr->length, fn); \
		break;

int gg_masq_manage_buffers(int minbufsize, struct gg_fd* f)
{
	if (minbufsize > f->bufsize)
	{
		if (gg_dynamic_buffers)
		{
			int newsize = f->bufsize << 1;
			while (minbufsize > newsize)
				newsize <<= 1;
			if (gg_masq_realloc_buffer(f, newsize))
				return 1;
			f->shrinkbuf = 0;
			return 2;
		}
		else
		{
			gg_condlog("warning: packet won't fit into buffer "
					"(fd=%d len=%d)", f->fd, minbufsize);
			return 1;
		}
	}
	else if (gg_dynamic_buffers)
	{
		if (minbufsize <= f->bufsize >> 1)
		{
			if (++(f->shrinkbuf) == GG_SHRINKBUF_THRESHOLD)
			{
				int newsize = f->bufsize >> 1;
				while (minbufsize <= newsize)
					newsize >>= 1;
				if (gg_masq_realloc_buffer(f, newsize<<1))
					return 0; /* non-fatal failure */
				f->shrinkbuf = 0;
				return 2;
			}
		}
		else
			f->shrinkbuf = 0;
	}
	return 0;
}

void gg_masq_hdr(struct gg_fd* f)
{
	struct gg_header* hdr = (struct gg_header*)f->buffer;
	int minbufsize = gg_max(hdr->length + sizeof(struct gg_header),
			f->readoffset);

	DEBUGP("fd=%d type=%d len=%d", f->fd, hdr->type, hdr->length);

	if (hdr->type < 0 || hdr->type > GG_MAX_PKT_TYPE || hdr->length < 0 ||
			hdr->length > GG_MAX_PKT_LEN)
		goto failure;

	switch (gg_masq_manage_buffers(minbufsize, f))
	{
		case 1:
			goto failure;
		case 2:
			hdr = (struct gg_header*)f->buffer;
	}

	if (f->flags & GG_FLAG_CLIENTCONN) /* outgoing packets */
		switch (hdr->type)
		{
			GG_MASQ_PKT(GG_LOGIN, struct gg_login, gg_masq_login)
			GG_MASQ_PKT(GG_LOGIN_EXT, struct gg_login_ext, gg_masq_login_ext)
			GG_MASQ_PKT(GG_LOGIN60, struct gg_login60, gg_masq_login60)

			default:
				gg_masq_cont(f, hdr->length, gg_masq_default);
		}
	else /* incoming packets */
		switch (hdr->type)
		{
			case GG_LOGIN_OK:
				if (hdr->length != 0)
					goto failure;
				DEBUGP("login ok on fd %d", f->fd);
				f->flags |= GG_FLAG_LOGGED;
				gg_masq_nextpkt(f);
				break;
				
			GG_MASQ_PKT(GG_NOTIFY_REPLY, struct gg_notify_reply,
					gg_masq_notify_reply)
			GG_MASQ_PKT(GG_NOTIFY_REPLY60, struct gg_notify_reply60,
					gg_masq_notify_reply60)
			GG_MASQ_PKT(GG_STATUS60, struct gg_status60, gg_masq_status60)

			default:
				gg_masq_cont(f, hdr->length, gg_masq_default);
		}
	return;
	
failure:
	gg_condlog("violation: type=%d len=%d", hdr->type, hdr->length);
	gg_proto_violated(f);
}

static void gg_masq_simple_inject(struct gg_fd* f, const char* buf, int len,
		void (*whatthen)(struct gg_fd*))
{
	gg_assert(f->writeoffset == 0 && f->readoffset == 0);
	gg_assert(len <= f->bufsize);
	DEBUGP("fd=%d len=%d", f->fd, len);
	memcpy(f->buffer, buf, len);
	f->writeoffset = len;
	f->readoffset = len;
	f->expectfn = whatthen;
	f->twin = f;
	f->state = GG_S_WRTCMPL;
}

static int gg_masq_inject(struct gg_fd* f, const char* buf, int len)
{
	if (GG_INJECT_BUFSIZE - f->injectlen < len)
	{
		DEBUGP("no storage for injection");
		return 1;
	}

	if (f->readoffset == 0 && f->state == GG_S_EXPECTING && 
			f->expectfn == gg_masq_hdr && f->bufsize >= len)
	{
		memcpy(f->buffer, buf, len);
		f->readoffset = f->writeoffset = len; /* just repeat */
	}
	else
	{
		memcpy(f->injectbuf + f->injectlen, buf, len);
		f->injectlen += len;
	}
	return 0;
}

static void gg_masq_devnull_buffer(struct gg_fd* f)
{
	f->writeoffset = f->readoffset = 0;
}

static void gg_masq_devnull_buffer_partially(struct gg_fd* f, int len)
{
	f->readoffset -= len;
	memmove(f->buffer, f->buffer + len, f->readoffset);
}

static void gg_masq_recv_dcctype(struct gg_fd* f)
{
	int* dt = (int*)f->buffer;
	struct gg_conn_expectation* exp;
	DEBUGP("dt=%d", *dt);

	switch (*dt)
	{
		case GG_DCC_WANT_FILE:
			if (!f->expby)
			{
				gg_condlog("WANT_FILE dcc method not implemented (fd=%d)", f->fd);
				goto failure;
			}
			
			gg_masq_devnull_buffer_partially(f, sizeof(int));
			f->state = GG_S_REPEATMODE;
			f->twin = f->expby;
			f->twin->twin = f;
			f->expby = NULL;
			f->twin->state = GG_S_REPEATMODE;
			gg_condlog("finalized dcc connection setup on fds %d:%d", f->fd,
					f->twin->fd);
			return;

		case GG_DCC_CATCH_FILE:
			if (f->expby) goto failure;

			if (!f->parent->twin)
			{
				DEBUGP("no parent->twin");
				goto failure;
			}

			gg_masq_devnull_buffer_partially(f, sizeof(int));
			f->state = GG_S_SUSPENDED;

			if (!(exp = (struct gg_conn_expectation*)
						malloc(sizeof(struct gg_conn_expectation))))
			{
				gg_error("warning: error allocating gg_conn_expectation");
				goto failure;
			}

			if (f->flags & GG_FLAG_DCCINCOMING)
			{
				struct gg_recv_msg_pkt msg;
				msg.hdr.type = GG_RECV_MSG;
				msg.hdr.length = sizeof(msg) - sizeof(msg.hdr);
				msg.sender = f->my_uin;
				msg.seq = 0;
				msg.time = 0;
				msg.msgclass = GG_CLASS_CTCP;
				msg.dunno = 2;
				if (gg_masq_inject(f->parent->twin, (char*)&msg, sizeof(msg)))
				{
					DEBUGP("error injecting data");
					goto freeexp;
				}
			}
			else if (f->flags & GG_FLAG_DCCOUTNATTED)
			{
				struct gg_send_msg_pkt msg;
				msg.hdr.type = GG_SEND_MSG;
				msg.hdr.length = sizeof(msg) - sizeof(msg.hdr);
				msg.recipient = f->your_uin;
				msg.seq = 0;
				msg.msgclass = GG_CLASS_CTCP;
				msg.dunno = 2;
				if (gg_masq_inject(f->parent, (char*)&msg, sizeof(msg)))
				{
					DEBUGP("error injecting data");
					goto freeexp;
				}
			}
			else goto freeexp;
			
			exp->twin = f;
			exp->my_uin = f->your_uin;
			exp->your_uin = f->my_uin;
			exp->owner = f;
			exp->timeout = time(NULL) + GG_EXPECTATION_TIMEOUT;
			f->cexp = exp;
			INIT_LIST_HEAD(&exp->list);
			list_add_tail(&exp->list, &gg_masq_cexps);
			DEBUGP("added expectation %d->%d", exp->my_uin, exp->your_uin);
			return;

		case GG_DCC_MANAGEMENT1:
			DEBUGP("aborting management connection");
			goto failure;

		default:
			gg_condlog("unsupported dcc method %d on fd %d", *dt, f->fd);
			goto failure;
	}

freeexp:
	free(exp);
failure:
	DEBUGP("failure, aborting");
	gg_abort_fd(f);
}

static void gg_masq_expect_dcctype(struct gg_fd* f)
{
	gg_assert(f->writeoffset == 0 && f->readoffset == 0);
	DEBUGP("fd=%d", f->fd);
	f->expectlen = sizeof(int);
	f->expectfn = gg_masq_recv_dcctype;
	f->twin = NULL;
	f->state = GG_S_EXPECTING;
}

static void gg_masq_setup_complex_dcc(struct gg_fd* f, int my_uin,
		int your_uin, struct gg_fd* parent, char type)
{
	f->parent = parent;
	f->your_uin = your_uin;
	f->my_uin = my_uin;
	f->flags |= type;
	gg_get_fd(parent);
	gg_masq_devnull_buffer(f);
	gg_masq_simple_inject(f, "UDAG", 4, gg_masq_expect_dcctype);
}

static int gg_masq_setup_simple_dcc(struct gg_fd* f, uint32_t ip,
		uint16_t port)
{
	struct gg_connect_info* info;
	pthread_t dummy;
	pthread_attr_t attr;

	if (!(info = (struct gg_connect_info*)malloc(sizeof(*info))))
		return 1;
	info->twin = f;
	info->nat_ip = ip;
	info->nat_port = port;
	f->state = GG_S_SUSPENDED;
	f->writeoffset = f->readoffset;
	gg_get_fd(f);
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	DEBUGP("spawning helper thread...");
	if (pthread_create(&dummy, &attr, gg_establish_dcc, info))
	{
		gg_put_fd(f);
		gg_close_fd(f);
		free(info);
		return 1;
	}
	return 0;
}

void gg_masq_dcc(struct gg_fd* f)
{
	struct gg_dcc_handshake* hs = (struct gg_dcc_handshake*)f->buffer;
	struct gg_uinh_entry* h_my, * h_your;
	struct gg_conn_expectation* e;

	if (hs->my_uin == 0) goto failure;
	if (hs->your_uin == 0) goto failure;

	list_for_each_entry(e, &gg_masq_cexps, list)
		if (e->my_uin == hs->my_uin && e->your_uin == hs->your_uin)
		{
			gg_condlog("expected connection caught");
			f->expby = e->twin;
			e->twin->cexp = NULL;
			gg_masq_devnull_buffer(f);
			gg_masq_simple_inject(f, "UDAG", 4, gg_masq_expect_dcctype);
			list_del(&e->list);
			free(e);
			return;
		}
	
	h_my = gg_uinh_find(hs->my_uin);
	h_your = gg_uinh_find(hs->your_uin);
	
	if (!h_my || !h_your)
	{
		DEBUGP("uin not found");
		goto failure;
	}

	if (h_your->fd)
	{
		gg_condlog("scheme (->in) caught, %d->%d", hs->my_uin, hs->your_uin);
		if (gg_cascade)
		{
			if (gg_masq_setup_simple_dcc(f, h_your->ip, h_your->port))
				goto failure;
		}
		else
			gg_masq_setup_complex_dcc(f, hs->my_uin, hs->your_uin, h_your->fd,
					GG_FLAG_DCCINCOMING);
		return;
	}
	else if (h_my->fd) /* our client is opening connection */
	{
		if (!h_your->ip)
		{
			gg_condlog("%d tried to connect to null-ip of %d",
					hs->my_uin, hs->your_uin);
			goto failure;
		}
		if (h_your->port < 3)
		{
			if (h_your->port != 1) /* our friend is natted? no. */
			{
				gg_condlog("dcc client %d requested by %d is not available",
						hs->my_uin, hs->your_uin);
				goto failure;
			}
			
			gg_condlog("scheme (in->outnatted) caught, %d->%d",
					hs->my_uin, hs->your_uin);
			gg_masq_setup_complex_dcc(f, hs->my_uin, hs->your_uin, h_my->fd,
					GG_FLAG_DCCOUTNATTED);
			return;
		}
		
		gg_condlog("initiated dcc connection (in->out) %lu -> %lu (%u.%u.%u.%u:%u) on fd %d",
				(unsigned long)hs->my_uin, (unsigned long)hs->your_uin,
				NIPQUAD(h_your->ip), h_your->port, f->fd);

		if (!gg_masq_setup_simple_dcc(f, h_your->ip, h_your->port))
			return;
	}

failure:
	gg_condlog("invalid conditions establishing dcc");
	gg_abort_fd(f);
}

void gg_masq_timeout_exps(void)
{
	time_t now = time(NULL);

	while (!list_empty(&gg_masq_cexps))
	{
		struct gg_conn_expectation* exp = list_entry(gg_masq_cexps.next,
				struct gg_conn_expectation, list);
		if (exp->timeout > now) break;
		gg_condlog("timed out expectation %d->%d", exp->my_uin, exp->your_uin);
		exp->twin->cexp = NULL;
		gg_abort_fd(exp->twin);
		list_del(&exp->list);
		free(exp);
	}
}

